# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# referenced from
# Library Name: torchtext
# Authors: torchtext authors and @sluks
# Date: 2020-07-18
# Link: https://pytorch.org/text/_modules/torchtext/data/metrics.html#bleu_score
from typing import Sequence

import torch
from torchmetrics.functional import bleu_score as _bleu_score

from pytorch_lightning.metrics.utils import deprecated_metrics, void


@deprecated_metrics(target=_bleu_score)
def bleu_score(
    translate_corpus: Sequence[str], reference_corpus: Sequence[str], n_gram: int = 4, smooth: bool = False
) -> torch.Tensor:
    """
    .. deprecated::
        Use :func:`torchmetrics.functional.bleu_score`. Will be removed in v1.5.0.
    """
    return void(translate_corpus, reference_corpus, n_gram, smooth)
